# Copyright 2024 Zoé Cassiopée Gauthier.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import contextlib
import functools
import json
import operator
import sqlite3
import time
import urllib.parse
from pathlib import Path
from typing import final

import requests
from rich.progress import Progress, TaskID

from pogo_scaled_estimators.utilities import format_move_name, format_pokemon_name

WEAKNESS = 1.6
DOUBLE_WEAKNESS = WEAKNESS * WEAKNESS


@final
class Calculator:
    def __init__(self, attacker_types: list[str]) -> None:
        self.db = sqlite3.connect("ase.db")
        # db.set_trace_callback(print)

        with contextlib.suppress(sqlite3.OperationalError):
            _ = self.db.execute(
                "CREATE TABLE estimators(defender, raid_tier, attacker, level, quick_move, charged_move, estimator, party)"
            )

        self.raids = self._pokebattler_resource("raids")
        self.pokemon = self._pokebattler_resource("pokemon")
        self.resists = self._pokebattler_resource("resists")
        self.moves = self._pokebattler_resource("moves")

        self.attacker_types = attacker_types

        self.progress = Progress()
        self._refresh_task: TaskID | None = None

    def _pokebattler_resource(self, name: str) -> dict:
        p = Path(f"./{name}.json")
        if not p.exists():
            response = requests.get(f"https://fight.pokebattler.com/{name}")
            with p.open(mode="wb") as fp:
                _ = fp.write(response.content)

        with p.open() as fp:
            return json.load(fp)

    def calculate(self, level: int = 40, party: int = 1) -> None:
        raid_bosses = self._raid_bosses()
        charged_moves = [
            move["moveId"] for move in self.moves["move"] if "type" in move and move["type"] in self.attacker_types
        ]
        res = self.db.execute(
            f"""SELECT DISTINCT(attacker) FROM estimators WHERE charged_move IN ("{'","'.join(charged_moves)}")"""
        )
        attackers = {row[0]: {"RAID_LEVEL_3": [], "RAID_LEVEL_5": [], "RAID_LEVEL_MEGA": []} for row in res.fetchall()}
        movesets = {}

        defenders = functools.reduce(operator.iconcat, raid_bosses.values(), [])
        res = self.db.execute(
            f"""
            SELECT e.raid_tier, e.defender, e.attacker, e.estimator / m.min_estimator, e.quick_move, e.charged_move
            FROM estimators e
            INNER JOIN (
                SELECT defender, MIN(estimator) AS min_estimator
                FROM estimators
                WHERE party = ? AND level = 40
                    AND attacker IN ("{'","'.join(attackers.keys())}")
                    AND defender IN ("{'","'.join(defenders)}")
                GROUP BY defender
            ) AS m ON e.defender = m.defender
            INNER JOIN (
                SELECT defender, attacker, MIN(estimator) as min_estimator
                FROM estimators
                WHERE party = ? AND level = ? AND charged_move IN ("{'","'.join(charged_moves)}")
                GROUP BY defender, attacker
            ) AS ms ON e.defender = ms.defender AND e.attacker = ms.attacker AND e.estimator = ms.min_estimator
            WHERE e.attacker IN ("{'","'.join(attackers.keys())}")
        """,
            (party, party, level),
        )
        for raid_tier, _, attacker, estimator, fast_move, charged_move in res.fetchall():
            if raid_tier == "RAID_LEVEL_MEGA_5":
                simplified_raid_tier = "RAID_LEVEL_MEGA"
            elif raid_tier == "RAID_LEVEL_ULTRA_BEAST":
                simplified_raid_tier = "RAID_LEVEL_5"
            else:
                simplified_raid_tier = raid_tier
            attackers[attacker][simplified_raid_tier].append(estimator)
            movesets[attacker] = (fast_move, charged_move)

        ase = {}
        for attacker, estimators in attackers.items():
            if not estimators["RAID_LEVEL_3"] or not estimators["RAID_LEVEL_5"] or not estimators["RAID_LEVEL_MEGA"]:
                continue
            ase = (
                0.15 * sum(estimators["RAID_LEVEL_3"]) / len(estimators["RAID_LEVEL_3"])
                + 0.50 * sum(estimators["RAID_LEVEL_5"]) / len(estimators["RAID_LEVEL_5"])
                + 0.35 * sum(estimators["RAID_LEVEL_MEGA"]) / len(estimators["RAID_LEVEL_MEGA"])
            )
            fast_move, charged_move = movesets[attacker]
            print(f"{attacker},{level},{ase},{fast_move},{charged_move}")

    def _raid_bosses(self) -> dict:
        raid_tiers = []
        raid_bosses = {}

        for raid_level in ["3", "5", "MEGA", "MEGA_5", "ULTRA_BEAST"]:
            tier = f"RAID_LEVEL_{raid_level}"
            raid_tiers.extend(
                [
                    tier,
                    f"{tier}_LEGACY",
                    f"{tier}_FUTURE",
                ]
            )
            raid_bosses[tier] = []

        for tier in filter(lambda tier: tier["tier"] in raid_tiers, self.raids["tiers"]):
            for boss in (raid["pokemon"] for raid in tier["raids"]):
                if boss.endswith("_FORM"):
                    continue
                boss_pokemon = next(filter(lambda mon: mon["pokemonId"] == boss, self.pokemon["pokemon"]))
                if ("candyToEvolve" in boss_pokemon or boss in ["SEADRA", "SEALEO"]) and boss not in [
                    "KELDEO",
                    "LUMINEON",
                    "MANAPHY",
                    "PHIONE",
                    "STUNFISK",
                    "TERRAKION",
                ]:
                    continue
                boss_types = (
                    boss_pokemon["type"],
                    boss_pokemon.get("type2", "POKEMON_TYPE_NONE"),
                )
                if any(self._is_weak(attacker_type, boss_types) for attacker_type in self.attacker_types):
                    raid_bosses[tier["info"]["guessTier"]].append(boss)

        return raid_bosses

    def _is_weak(self, attacker_type: str, defender_types: tuple[str, str]) -> bool:
        pokemon_types = list(self.resists.keys())
        defender_type_indices = (
            pokemon_types.index(defender_types[0]),
            pokemon_types.index(defender_types[1]),
        )
        attack_resist = (
            self.resists[attacker_type][defender_type_indices[0]]
            * self.resists[attacker_type][defender_type_indices[1]]
        )

        # Check for double weakness.
        if attack_resist >= DOUBLE_WEAKNESS:
            return True

        # Checkout for single weakness if not double weak to anything else.
        any_double_weaknesses = any(
            r[defender_type_indices[0]] * r[defender_type_indices[1]] >= DOUBLE_WEAKNESS for r in self.resists.values()
        )
        if not any_double_weaknesses and attack_resist >= WEAKNESS:
            return True

        return False

    def load_estimators(self, tier: str, defender: str, level: int, party: int) -> None:
        base_url = "https://fight.pokebattler.com"
        query_string = {
            "sort": "ESTIMATOR",
            "weatherCondition": "NO_WEATHER",
            "dodgeStrategy": "DODGE_REACTION_TIME",
            "aggregation": "AVERAGE",
            "includeLegendary": "true",
            "includeShadow": "true",
            "includeMegas": "true",
            "primalAssistants": "",
            "numParty": str(party),
        }
        if tier != "RAID_LEVEL_3":
            query_string["friendshipLevel"] = "FRIENDSHIP_LEVEL_4"
        url = f"{base_url}/raids/defenders/{defender}/levels/{tier}/attackers/levels/{level}/strategies/CINEMATIC_ATTACK_WHEN_POSSIBLE/DEFENSE_RANDOM_MC?{urllib.parse.urlencode(query_string, doseq=True)}"
        response = requests.get(url)
        for attacker in response.json()["attackers"][0]["randomMove"]["defenders"]:
            for attacker_moves in attacker["byMove"]:
                res = self.db.execute(
                    "SELECT estimator FROM estimators WHERE defender=? AND attacker=? AND level=? AND quick_move=? AND charged_move=? AND party=?",
                    (
                        defender,
                        attacker["pokemonId"],
                        level,
                        attacker_moves["move1"],
                        attacker_moves["move2"],
                        party,
                    ),
                )
                estimator = res.fetchone()
                if estimator is not None:
                    self.progress.console.log(
                        f'{format_pokemon_name(attacker["pokemonId"])} ({format_move_name(attacker_moves["move1"])}/{format_move_name(attacker_moves["move2"])}): {estimator[0]:.2f} (cached)'
                    )
                    continue

                _ = self.db.execute(
                    "INSERT INTO estimators(defender, raid_tier, attacker, level, quick_move, charged_move, estimator, party) VALUES(?, ?, ?, ?, ?, ?, ?, ?)",
                    (
                        defender,
                        tier,
                        attacker["pokemonId"],
                        level,
                        attacker_moves["move1"],
                        attacker_moves["move2"],
                        attacker_moves["result"]["estimator"],
                        party,
                    ),
                )
                self.progress.console.log(
                    f'{format_pokemon_name(attacker["pokemonId"])} ({format_move_name(attacker_moves["move1"])}/{format_move_name(attacker_moves["move2"])}): {attacker_moves["result"]["estimator"]:.2f}'
                )

            if self._refresh_task is not None:
                self.progress.update(self._refresh_task, advance=1)
            self.db.commit()

    def refresh(self, level: int = 40, party: int = 1) -> None:
        with self.progress:
            total_defenders = sum(len(defenders) for defenders in self._raid_bosses().values())
            self._refresh_task = self.progress.add_task("Working...", total=(total_defenders * 30))
            for tier, defenders in self._raid_bosses().items():
                for defender in defenders:
                    self.progress.update(self._refresh_task, description=f"vs {format_pokemon_name(defender)}...")
                    try:
                        self.load_estimators(tier, defender, level, party)
                    except json.decoder.JSONDecodeError:
                        time.sleep(30)
                        self.load_estimators(tier, defender, level, party)
                    time.sleep(1)
